import binascii
import os
import time
from dataclasses import dataclass

from cryptography.hazmat.primitives.kdf.scrypt import Scrypt
from nanoid import generate

from labfunctions.defaults import NANO_ID_ALPHABET


def generate_random(size=10, strategy="nanoid", alphabet=NANO_ID_ALPHABET) -> str:
    """Default URLSafe id"""
    if strategy == "nanoid":
        return generate(alphabet=alphabet, size=size)
    raise NotImplementedError("Strategy %s not implemented", strategy)


@dataclass
class Hash96:
    """A custom id generator of 96 bits long, compose as:
        [ 41 bits -> TS ][55 bits of a random number]
    in this way the generated ids are naturally ordered by time.
    """

    id_int: int
    id_hex: str

    @classmethod
    def time_random_string(cls):
        """
        It generates a random string based on a time epoch
        an a random number.
        It's a 96 bit long

        """
        # 41 bits timestamp with a custom epoch
        ts = int(time.time())
        # 55 bits a random number
        num = int.from_bytes(os.urandom(90), "big") % 2**55
        _int_id = (ts << 55) | (num << 0)
        _bytes = _int_id.to_bytes(12, "big")
        _bytes_id = binascii.hexlify(_bytes).decode()

        return cls(id_int=_int_id, id_hex=_bytes_id)


class PasswordScript:
    """
    https://cryptography.io/en/latest/hazmat/primitives/key-derivation-functions/#cryptography.hazmat.primitives.kdf.scrypt.Scrypt
    """

    def __init__(self, salt: bytes, n=2**14, r=8, p=1):
        self.n = n
        self.r = r
        self.p = p
        self.salt = salt

    def _kdf(self):
        kdf = Scrypt(
            salt=self.salt,
            length=32,
            n=self.n,
            r=self.r,
            p=self.p,
        )
        return kdf

    def encrypt(self, pass_: str) -> bytes:
        kdf = self._kdf()

        return kdf.derive(pass_.encode("utf-8"))

    def verify(self, pass_: str, key: bytes) -> bool:
        kdf = self._kdf()
        try:
            kdf.verify(pass_.encode("utf-8"), key)
            return True
        except:
            return False



from __future__ import annotations

from copy import deepcopy
from typing import Any, Callable, Dict, List, Optional, Set

import jwt
from pydantic.error_wrappers import ValidationError

from labfunctions import defaults
from labfunctions.security import scopes
from labfunctions.types.config import SecuritySettings
from labfunctions.types.security import JWTConfig, JWTResponse
from labfunctions.utils import get_class

from .base import AuthSpec, TokenStoreSpec
from .errors import AuthValidationFailed
from .utils import get_delta, open_keys


class Auth(AuthSpec):
    def __init__(self, conf: JWTConfig, store: Optional[TokenStoreSpec] = None):
        """
        It is a wrapper around jwt which produces jwt tokens.
        By default it will add a "exp" claim, other claims.

        Standard claims could be configurated from JWTConfing or when passing
        a payload to encode. In that case if both configurations exists, it will
        prioritize the payload configuration.
        """
        self.conf = conf
        self.store = store

    def _get_secret_encode(self):
        """because jwt allows the use of a simple secret or a pairs of keys
        this function will look at the configuration to determine a secret to be used
        """
        if self.conf.keys:
            _secret = self.conf.keys.private
        else:
            _secret = self.conf.secret
        return _secret

    def _get_secret_decode(self):
        if self.conf.keys:
            _secret = self.conf.keys.public
        else:
            _secret = self.conf.secret
        return _secret

    def _build_payload(self, payload: Dict[str, Any], exp=None, iss=None, aud=None):
        _payload = deepcopy(payload)
        if not exp:
            exp = get_delta(self.conf.exp_min)
        _iss = iss or self.conf.issuer
        if _iss:
            _payload.update({"iss": _iss})

        _aud = aud or self.conf.audience
        if _aud:
            _payload.update({"aud": _aud})

        _payload.update({"exp": exp})
        return _payload

    def encode(self, payload: Dict[str, Any], exp=None, iss=None, aud=None) -> str:
        """Encode a payload into a JWT Token.

        Some standards claims like exp, iss and aud could be
        overwritten using this params.

        :param payload: a dictionary with any k/v pairs to add
        :param exp:  the “exp” (expiration time) claim identifies the expiration
        time on or after which the JWT MUST NOT be accepted for processing.
        if date or int is given it will overwrite the default configuration.
        :param iss: the “iss” (issuer) claim identifies the principal that issued the JWT.
        :param aud: The “aud” (audience) claim identifies the recipients that the JWT
        is intended for.
        """
        _secret = self._get_secret_encode()

        final = self._build_payload(payload, exp, iss, aud)

        encoded = jwt.encode(
            final,
            _secret,
            algorithm=self.conf.alg,
        )
        return encoded

    def decode(
        self, encoded, verify_signature=True, verify_exp=True, iss=None, aud=None
    ) -> Dict[str, Any]:
        _secret = self._get_secret_decode()

        _iss = iss or self.conf.issuer
        _aud = aud or self.conf.audience

        return jwt.decode(
            encoded,
            _secret,
            options={
                "verify_signature": verify_signature,
                "verify_exp": verify_exp,
                "require": self.conf.requires_claims,
            },
            aud=_aud,
            iss=_iss,
            algorithms=[self.conf.alg],
        )

    def validate(
        self,
        token: str,
        required_scopes: Optional[List[str]],
        require_all=True,
        iss=None,
        aud=None,
    ) -> Dict[str, Any]:
        try:
            decoded = self.decode(token, iss=iss, aud=aud)
            if required_scopes:
                user_scopes: List[str] = decoded["scopes"]
                valid = scopes.validate(
                    required_scopes, user_scopes, require_all=require_all
                )
                if not valid:
                    raise AuthValidationFailed()
        except jwt.InvalidTokenError as e:
            raise AuthValidationFailed()

        return decoded

    async def store_refresh_token(self, username: str, ttl=None) -> str:
        ttl = ttl or self.conf.ttl_refresh_token
        refresh = self.store.generate()
        await self.store.put(f"{username}.{refresh}", username, ttl=ttl)

        return refresh

    async def validate_refresh_token(self, access_token, refresh_token) -> bool:
        is_valid = False
        decoded = self.decode(access_token, verify_exp=False)
        value = await self.store.get(f"{decoded['usr']}.{refresh_token}")
        if decoded["usr"] == value:
            is_valid = True
        return is_valid

    async def refresh_token(self, access_token, refresh_token) -> JWTResponse:
        is_valid = await self.validate_refresh_token(access_token, refresh_token)
        if is_valid:
            decoded = self.decode(access_token, verify_exp=False)
            await self.store.delete(f"{decoded['usr']}.{refresh_token}")

            _new_refresh = await self.store_refresh_token(decoded["usr"])
            _new_tkn = self.encode(decoded)
            new_jwt = JWTResponse(access_token=_new_tkn, refresh_token=_new_refresh)
            return new_jwt
        raise AuthValidationFailed()


def auth_from_settings(
    settings: SecuritySettings, store: Optional[TokenStoreSpec] = None
) -> AuthSpec:
    """Intiliazie a `Auth` based on settings."""
    AuthClass = get_class(settings.AUTH_CLASS)
    keys = open_keys(settings.JWT_PUBLIC, settings.JWT_PRIVATE)
    conf = JWTConfig(
        alg=settings.JWT_ALG,
        exp_min=settings.JWT_EXP,
        keys=keys,
        issuer=settings.JWT_ISS,
        audience=settings.JWT_AUD,
        requires_claims=settings.JWT_REQUIRES_CLAIMS,
    )
    return AuthClass(conf, store=store)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['labfunctions',
 'labfunctions.client',
 'labfunctions.cluster',
 'labfunctions.cmd',
 'labfunctions.conf',
 'labfunctions.conf.templates',
 'labfunctions.control_plane',
 'labfunctions.db',
 'labfunctions.db.migrations',
 'labfunctions.db.migrations.versions',
 'labfunctions.errors',
 'labfunctions.executors',
 'labfunctions.io',
 'labfunctions.managers',
 'labfunctions.notebooks',
 'labfunctions.runtimes',
 'labfunctions.security',
 'labfunctions.types',
 'labfunctions.web']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=3.0.3,<4.0.0',
 'PyJWT>=2.1.0,<2.2.0',
 'PyYAML>=6.0,<7.0',
 'aiofiles>=0.8.0,<0.9.0',
 'aioredis[hiredis]>=2.0.1,<3.0.0',
 'aiosqlite>=0.17.0,<0.18.0',
 'asyncssh>=2.10.0,<3.0.0',
 'click>=8.0.1,<9.0.0',
 'cloudpickle>=2.0.0,<3.0.0',
 'cryptography>=36.0.1,<37.0.0',
 'dateparser>=1.1.0,<2.0.0',
 'docker>=5.0.3,<6.0.0',
 'httpx<0.22.0',
 'ipykernel>=6.9.1,<7.0.0',
 'ipython>=8.1.1,<9.0.0',
 'jupytext>=1.13.0,<2.0.0',
 'loky>=3.0.0,<4.0.0',
 'nanoid>=2.0.0,<3.0.0',
 'nbconvert>=6.2.0,<7.0.0',
 'papermill>=2.3.4,<3.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'pytz>=2021.1,<2022.0',
 'rich>=12.0.0,<13.0.0',
 'tenacity>=8.0.1,<9.0.0',
 'tqdm>=4.62.3,<5.0.0']

extras_require = \
{'cloud': ['smart-open[s3,gcs]>=6.0.0,<7.0.0', 'apache-libcloud>=3.5.1,<4.0.0'],
 'server': ['uvloop>=0.16.0,<0.17.0',
            'sanic>=21.6.2,<22.0.0',
            'sanic-openapi>=21.6.1,<22.0.0',
            'sanic-ext>=21.9.0,<22.0.0',
            'SQLAlchemy[asyncio]>=1.4.26,<2.0.0',
            'SQLAlchemy-serializer>=1.4.1,<2.0.0',
            'psycopg2-binary>=2.9.1,<3.0.0',
            'asyncpg>=0.24.0,<0.25.0',
            'alembic>=1.6.5,<2.0.0',
            'rq-scheduler>=0.11.0,<0.12.0',
            'rq>=1.10.0,<2.0.0']}

entry_points = \
{'console_scripts': ['lab = labfunctions.cli:cli']}

setup_kwargs = {
    'name': 'labfunctions',
    'version': '0.8.0a7',
    'description': 'Schedule parameterized notebooks programmatically using cli or a REST API',
    'long_description': "# LabFunctions\n\n[![labfunctions](https://github.com/nuxion/labfunctions/actions/workflows/main.yaml/badge.svg)](https://github.com/nuxion/labfunctions/actions/workflows/main.yaml)\n[![readthedocs](https://readthedocs.org/projects/labfunctions/badge/?version=latest)](https://labfunctions.readthedocs.io/en/latest/)\n[![PyPI - Version](https://img.shields.io/pypi/v/labfunctions)](https://pypi.org/project/labfuncions/)\n[![PyPI - Format](https://img.shields.io/pypi/format/labfunctions)](https://pypi.org/project/labfunctions/)\n[![PyPI - Status](https://img.shields.io/pypi/status/labfunctions)](https://pypi.org/project/labfunctions/)\n[![Docker last](https://img.shields.io/docker/v/nuxion/labfunctions/0.7.0)](https://hub.docker.com/r/nuxion/labfunctions/tags)\n[![codecov](https://codecov.io/gh/nuxion/labfunctions/branch/main/graph/badge.svg?token=F025Y1BF9U)](https://codecov.io/gh/nuxion/labfunctions)\n\n\n## :books: Description \n\nLabFunctions empowers different data roles to put notebooks into production, simplifying the time required to do so. It enables people to go from a data exploration instance to an entirely project deployed in production, using the same notebooks files made by a data scientist, analyst or whatever role working with data in an iterative way.\n\nLabFunctions is a library and a service that allows you to run parametrized notebooks in a distributed way.  \n\nA Notebook could be launched remotly on demand, or could be scheduled by intervals or using cron syntax.\n\nInternally it uses [Sanic](https://sanicframework.org) as web server, [papermill](https://papermill.readthedocs.io/en/latest/) as notebook executor, an [RQ](https://python-rq.org/)\nfor task distributions and coordination. \n\n:tada: Demo :tada: \n\n:floppy_disk: [Example project](\nhttps://github.com/nuxion/nbwf-demo2)\n\n\n## :telescope: Philosophy\n\nLabFunctions it insn't a complete MLOps solution.\n\nWe try hard to simply and expose the right APIs to the user for the part of scheduling notebooks with reproducibility in mind.\n\nWe also try to give the user the same freedom that lego tiles can gives, but we are opinated in the sense that all code, artifact, dependency and environment should be declareted\n\nWith this point of view, then: \n\n1) Git is neccesary :wink:\n2) Docker is necessary for environment reproducibility. \n3) Although you can push not versioned code,  versioning is almost enforced, and is always a good practice in software development\n\nThe idea comes from a [Netflix post](https://netflixtechblog.com/notebook-innovation-591ee3221233) which suggest using notebooks like an interface or a some kind of DSL to orchestrate different workloads like Spark and so on. But it also could be used to run entire process: like training a model, crawlings sites, performing etls, and so on. \n\nThe benefits of this approach is that notebooks runned could be stored and inspected for good or for bad. If something fails, is easy to run in a classical way: cell by cell. \n\nThe last point to clarify and it could challange the common sense or the way that we are used to use Jupyter's Notebooks, is that each notebook is more like a function definition with inputs and outputs, so a notebook potentially could be used for different purposes, hence the name of **workflow**, and indeed this idea is common in the data space. Then a workflow will be a notebook with params defined to be used anytime that a user wants, altering or not the parameters sent. \n\n\n## :nut_and_bolt: Features\n\n- Define a notebook like a function, and execute it on demand or scheduling it\n- Automatic Dockerfile generation. A project should share a unique environment but could use different versions of the same environment\n- Execution History, Notifications to Slack or Discord.\n- Cluster creation applying scaling policies by idle time or/and enqueued items\n\n## Cluster options\n\nIt is possible to run different cluster configurations with custom auto scalling policies\n\n![GPU CLUSTER DEMO](https://media.giphy.com/media/OnhmnYiCJpe2FsTmaP/giphy.gif)\n\nInstances inside a cluster could be created manually or automatically\n\nSee a simple demo of a gpu cluster creation\n\n[https://www.youtube.com/watch?v=-R7lJ4dGI9s](https://www.youtube.com/watch?v=-R7lJ4dGI9s)\n\n\n## Installation\n\n### Server\n\n### Docker-compose\n\nThe project provides a [docker-compose.yaml](./docker-compose.yaml) file as en example. \n\n:construction: Note :construction:\n\nBecause **NB Workflows** will spawn docker instance for each workload, the installation inside docker containers could be tricky. \nThe most difficult part is the configuration of the worker that needs access to the [docker socket](https://docs.docker.com/engine/reference/commandline/dockerd/#daemon-socket-option).\n\nA [Dockerfile](./Dockerfile) is provided for customization of uid and guid witch should match with the local environment. A second alternative is expose the docker daemon through HTTP, if that is the case a `DOCKER_HOST` env could be used, see [docker client sdk](https://docker-py.readthedocs.io/en/stable/client.html)\n\n\n```\ngit clone https://github.com/nuxion/labfunctions\ncd labfunctions\n```\n\nThe next step is intializing the database and creating a user (*please review the script first*):\n\n```\ndocker-compose postgres up -d \n./scripts/initdb_docker.sh\n```\nNow you can start everything else:\n```\ndocker-compose up -d \n```\n\n### Without docker\n\n```\npip install nb-workflows[server]==0.6.0\n```\n\nfirst terminal:\n\n```\nexport NB_SERVER=True\nnb manager db upgrade\nnb manager users create\nnb web --apps workflows,history,projects,events,runtimes\n```\n\nsecond terminal:\n\n```\nnb rqworker -w 1 -q control,mch.default\n```\n\nBefore all that, redis postgresql and the [nginx in webdav mode](./fileserver.conf) should be configurated\n\n### Client\n\nClient: \n\n```\npip install nb-workflows==0.6.0\nnb startporject .\n```\n\n\n## :earth_americas: Roadmap\n\nSee [Roadmap](/ROADMAP.md) *draft*\n\n## :post_office: Architecture\n\n![labfunctions architecture](/docs/platform-workflows.jpg)\n\n\n## :bookmark_tabs: References & inspirations\n- [Notebook Innovation - Netflix](https://netflixtechblog.com/notebook-innovation-591ee3221233)\n- [Tensorflow metastore](https://www.tensorflow.org/tfx/guide/mlmd)\n- [Maintainable and collaborative pipelines](https://blog.jupyter.org/ploomber-maintainable-and-collaborative-pipelines-in-jupyter-acb3ad2101a7)\n- [The magic of Merlin](https://shopify.engineering/merlin-shopify-machine-learning-platform)\n- [Scale aware approach](https://queue.acm.org/detail.cfm?id=3025012)\n\n\n",
    'author': 'nuxion',
    'author_email': 'nuxion@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/nuxion/labfunctions',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



from sanic.exceptions import SanicException
from sanic.exceptions import Unauthorized as SanicUnauthorized


class SanicJWTException(SanicException):
    pass


class InvalidToken(SanicJWTException):
    pass


class WebAuthFailed(SanicJWTException):
    status_code = 401
    quiet = True

    def __init__(self, message="Authentication failed.", **kwargs):
        super().__init__(message, **kwargs)


class MissingAuthorizationHeader(SanicJWTException):
    status_code = 400

    def __init__(self, message="Authorization header not present.", **kwargs):
        super().__init__(message, **kwargs)


class Unauthorized(SanicJWTException, SanicUnauthorized):
    def __init__(self, message="Auth required.", **kwargs):
        super().__init__(message, scheme="Bearer", **kwargs)


class AuthValidationFailed(Exception):
    pass



from typing import Any, Dict, List, Optional

from pydantic import BaseSettings, RedisDsn

from labfunctions.defaults import EXECID_LEN, PROJECTID_MIN_LEN, WFID_LEN


class SecuritySettings(BaseSettings):
    JWT_PUBLIC: str
    JWT_PRIVATE: str
    JWT_ALG: str = "ES512"
    JWT_EXP: int = 30  # 30 minutes
    # JWT_PUBLIC: str = ".secrets/ecdsa.pub.pem"
    # JWT_PRIVATE: str = ".secrets/ecdsa.priv.pem"
    JWT_REQUIRES_CLAIMS: List[str] = ["exp"]
    JWT_SECRET: Optional[str] = None
    JWT_ISS: Optional[str] = None
    JWT_AUD: Optional[str] = None
    REFRESH_TOKEN_TTL: int = 3600 * 168  # 7 days
    TOKEN_STORE_URL: Optional[str] = None
    AUTH_SALT: str = "changeit"
    AUTH_ALLOW_REFRESH: bool = True
    AUTH_CLASS = "labfunctions.security.authentication.Auth"
    AUTH_FUNCTION = "labfunctions.managers.users_mg.authenticate"

    class Config:
        env_prefix = "LF_"


class ServerSettings(BaseSettings):
    # Security
    AGENT_TOKEN: str
    AGENT_REFRESH_TOKEN: str

    # Services
    WORKFLOW_SERVICE: str
    SQL: str
    ASQL: str
    AGENT_TOKEN_EXP: int = (60 * 60) * 12

    # Folders:
    BASE_PATH: str
    DOCKER_REGISTRY: Optional[str] = None

    SECURITY: Optional[SecuritySettings] = None

    DEV_MODE: bool = False
    WEB_REDIS: Optional[RedisDsn] = None
    RQ_REDIS: Optional[RedisDsn] = None
    CONTROL_QUEUE: str = "default.control"
    BUILD_QUEUE: str = "default.build"

    # ids generations
    EXECID_LEN: int = EXECID_LEN
    PROJECTID_LEN: int = PROJECTID_MIN_LEN
    WFID_LEN: int = WFID_LEN

    # eventes
    EVENTS_BLOCK_MS: int = 10 * 1000
    EVENTS_STREAM_TTL_SECS: int = 60 * 60

    # docker
    DOCKER_UID: str = "1000"
    DOCKER_GID: str = "997"

    # cluster
    AGENT_DATA_FOLDER: str = ".worker_data/"
    CLUSTER_SSH_KEY_USER: str = "op"
    CLUSTER_SSH_PUBLIC_KEY: Optional[str] = None
    CLUSTER_SPEC: str = "scripts/local_clusters.yaml"
    AGENT_HOMEDIR: str = "/home/op"
    AGENT_ENV_FILE: str = ".env.dev.docker"
    AGENT_HEARTBEAT_CHECK: int = 60 * 5
    AGENT_HEARTBEAT_TTL: int = 80 * 3

    # Logs:
    LOGLEVEL: str = "INFO"
    LOGCONFIG: Dict[str, Any] = {}
    DEBUG: bool = False

    PROJECTS_STORE_CLASS_ASYNC = "labfunctions.io.kv_local.AsyncKVLocal"
    PROJECTS_STORE_CLASS_SYNC = "labfunctions.io.kv_local.KVLocal"
    PROJECTS_STORE_BUCKET = "nbworkflows"
    EXT_KV_LOCAL_ROOT: Optional[str] = None
    EXT_KV_FILE_URL: Optional[str] = None

    SETTINGS_MODULE: Optional[str] = None
    DNS_IP_ADDRESS: str = "8.8.8.8"

    class Config:
        env_prefix = "LF_"


class ClientSettings(BaseSettings):
    WORKFLOW_SERVICE: str
    PROJECTID: str

    BASE_PATH: str
    LOGLEVEL: str = "INFO"
    LOGCONFIG: Dict[str, Any] = {}
    DEBUG: bool = False
    LOCAL: bool = False
    NBVARS: Optional[str] = None
    AGENT_TOKEN: Optional[str] = None
    AGENT_REFRESH_TOKEN: Optional[str] = None

    PROJECT_NAME: Optional[str] = None
    PROJECTID_LEN: int = 10
    SLACK_BOT_TOKEN: Optional[str] = None
    SLACK_CHANNEL_OK: Optional[str] = None
    SLACK_CHANNEL_FAIL: Optional[str] = None
    DISCORD_OK: Optional[str] = None
    DISCORD_FAIL: Optional[str] = None

    PROJECTS_STORE_CLASS = "labfunctions.io.kv_local.KVLocal"
    PROJECTS_STORE_BUCKET = "nbworkflows"
    EXT_KV_LOCAL_ROOT: Optional[str] = None
    EXT_KV_FILE_URL: Optional[str] = None

    SETTINGS_MODULE: Optional[str] = None
    # DOCKER_COMPOSE: Dict[str, Any] = None

    class Config:
        env_prefix = "LF_"



import importlib
import logging
import os
import sys

from .utils import load_client


def _get_level(level):
    return getattr(logging, level)


# singleton pattern
# https://github.com/samuelcolvin/pydantic/issues/586
settings = load_client()



from typing import List, Optional

from pydantic import BaseModel


class WFCreateRsp(BaseModel):
    status_code: int
    alias: str
    msg: Optional[str] = None
    wfid: Optional[str] = None


class WFPushRsp(BaseModel):
    created: Optional[List[WFCreateRsp]] = []
    errors: Optional[List[WFCreateRsp]] = []



from functools import wraps
from inspect import isawaitable
from typing import Callable, List, Optional

from sanic import Request, Sanic, json, text

from labfunctions import defaults
from labfunctions.types.config import SecuritySettings
from labfunctions.types.security import JWTConfig, JWTResponse, UserLogin
from labfunctions.utils import get_class

from .base import AuthSpec
from .errors import AuthValidationFailed, MissingAuthorizationHeader, WebAuthFailed


def get_auth(request: Request) -> AuthSpec:
    """a shortcut to get the Auth object from a web context"""

    current_app: Sanic = Sanic.get_app(request.app.name)

    return current_app.ctx.auth


def get_authenticate(request: Request) -> Callable:
    """a shortcut to get the Auth object from a web context"""
    current_app: Sanic = Sanic.get_app(request.app.name)

    return current_app.ctx.authenticate


def protected(scopes: Optional[List[str]] = None, require_all=True):
    """verify a token from a request.
    Optionally if a list of scopes is given then will check that scopes
    with the scopes provided by the token.

    :param scopes: a list of scopes
    :param required_all: if true it will check that the all the names provided
    match with the required.

    """

    def decorator(f):
        @wraps(f)
        async def decorated_function(request, *args, **kwargs):
            auth = get_auth(request)
            token = request.token
            if not token:
                raise MissingAuthorizationHeader()
            try:
                decoded = auth.validate(token, scopes, require_all)
                request.ctx.token_data = decoded
                response = f(request, *args, **kwargs)
                if isawaitable(response):
                    response = await response
            except AuthValidationFailed:
                raise WebAuthFailed()

            return response

        return decorated_function

    return decorator


def auth_error_handler(request, exception):
    return json({"msg": "Authentication failed"}, status=401)


def sanic_init_auth(app: Sanic, auth: AuthSpec, settings: SecuritySettings):
    app.ctx.auth = auth
    app.config.AUTH_SALT = settings.AUTH_SALT
    app.config.AUTH_ALLOW_REFRESH = settings.AUTH_ALLOW_REFRESH
    app.ctx.authenticate = get_class(settings.AUTH_FUNCTION)

    app.error_handler.add(WebAuthFailed, auth_error_handler)
    app.error_handler.add(MissingAuthorizationHeader, auth_error_handler)



import importlib
import sys
from getpass import getpass

import click
from alembic import command
from alembic.config import Config as AlembicConfig
from rich import print_json
from rich.console import Console
from rich.prompt import Confirm, Prompt

from labfunctions.conf import load_server
from labfunctions.db.sync import SQL
from labfunctions.managers import projects_mg, users_mg
from labfunctions.security import auth_from_settings
from labfunctions.security.redis_tokens import RedisTokenStore
from labfunctions.server import create_web_redis
from labfunctions.types.user import UserOrm
from labfunctions.utils import run_sync

settings = load_server()
console = Console()


@click.group(name="manager")
def managercli():
    """
    Managment tasks like user creation and db init
    """
    pass


def alembic_ugprade(dburi, to="head"):
    alembic_cfg = AlembicConfig("labfunctions/db/alembic.ini")
    alembic_cfg.set_main_option("sqlalchemy.url", dburi)
    command.upgrade(alembic_cfg, to)


@managercli.command()
@click.option("--sql", "-s", default=settings.SQL, help="SQL Database")
@click.argument("action", type=click.Choice(["create", "drop", "upgrade"]))
def db(sql, action):
    """Create or Drop tables from a database"""
    db = SQL(sql)
    settings.SQL = sql
    # auth_mod = importlib.import_module("labfunctions.auth.models")
    wf_mod = importlib.import_module("labfunctions.models")

    if action == "create":
        db.create_all()
        click.echo("Created...")
    elif action == "drop":
        db.drop_all()
        click.echo("Droped...")
    elif action == "upgrade":
        alembic_ugprade(sql)
    else:
        click.echo("Wrong param...")


@managercli.command()
@click.option("--sql", "-s", default=settings.SQL, help="SQL Database")
@click.option(
    "--superuser", "-S", is_flag=True, default=False, help="User as supersuer"
)
@click.option("--scopes", default="user", help="User scopes")
@click.option("--username", "-u", default=None, help="Username")
@click.argument(
    "action", type=click.Choice(["create", "disable", "reset", "change-scopes"])
)
def users(sql, superuser, scopes, username, action):
    """Manage users"""
    db = SQL(sql)
    if action == "create":
        name = Prompt.ask("Username")
        email = Prompt.ask("Email (optional)", default=None)
        password = getpass("Password: ")
        repeat = getpass("Password (repeat): ")
        if password != repeat:
            console.print("[bold red]Paswords doesn't match[/]")
            sys.exit(-1)

        S = db.sessionmaker()
        with S() as session:
            obj = UserOrm(
                username=name,
                # password=key,
                email=email,
                is_active=True,
                scopes=scopes,
                is_superuser=superuser,
            )
            user = users_mg.create(
                session, obj, password=password, salt=settings.SECURITY.AUTH_SALT
            )
            session.commit()
        console.print(f"[bold magenta]Congrats!! user {name} created")

    elif action == "disable":
        S = db.sessionmaker()
        with S() as session:
            u = users_mg.disable_user(session, username)
            session.commit()
            if u:
                click.echo(f"{username} disabled")
            else:
                click.echo(f"{username} not found")

    # elif action == "change-scopes":
    #     S = db.sessionmaker()
    #     with S() as session:
    #         u = users_mg.change_scopes(session, username, scopes)
    #         session.commit()
    #         if u:
    #             click.echo(f"{username} scopes changed")
    #         else:
    #             click.echo(f"{username} not found")

    elif action == "reset":
        name = Prompt.ask("Username")
        _p = getpass("Password: ")
        S = db.sessionmaker()
        with S() as session:
            changed = users_mg.change_pass(
                session, name, _p, salt=settings.SECURITY.AUTH_SALT
            )
            session.commit()
        if changed:
            console.print("[bold magenta]Pasword changed[/]")
        else:
            console.print("[bold red]User may not exist [/]")
    else:
        console.print("[red bold]Wrongs params[/]")


@managercli.command()
@click.option("--sql", "-s", default=settings.SQL, help="SQL Database")
@click.option("--scopes", default="agent:r:w", help="agent scopes")
@click.option("--username", "-u", default=None, help="Agent username")
@click.option("--admin", "-A", default=False, is_flag=True, help="Agent as admin")
@click.option("--exp", "-e", default=30, help="Expire time")
@click.argument("action", type=click.Choice(["create", "get-token", "delete"]))
def agent(sql, scopes, username, action, admin, exp):
    db = SQL(sql)
    store = RedisTokenStore(settings.WEB_REDIS)
    auth = auth_from_settings(settings.SECURITY, store=store)

    if action == "create":
        S = db.sessionmaker()
        with S() as session:
            um = run_sync(
                projects_mg.create_agent, session, scopes=scopes, is_admin=admin
            )
            jwt = run_sync(users_mg.get_jwt_token, auth, um, exp)
            session.commit()
            print_json(
                data={
                    "username": um.username,
                    "scopes": um.scopes.split(","),
                    "jwt": jwt.dict(),
                }
            )
    elif action == "get-token":
        if not username:
            console.print("[bold red]A username should be given[/]")
            return
        S = db.sessionmaker()
        with S() as session:
            um = users_mg.get_user(session, username)
            if not um:
                console.print("[bold red]Username not found[/]")
                return
            jwt = run_sync(users_mg.get_jwt_token, auth, um, exp)
            print_json(
                data={
                    "username": um.username,
                    "scopes": um.scopes.split(","),
                    "jwt": jwt.dict(),
                }
            )

    elif action == "delete":
        if not username:
            console.print("[bold red]A username should be given[/]")
            return
        S = db.sessionmaker()
        with S() as session:
            users_mg.delete_user(session, username)
            session.commit()
        console.print("[bold green]Agent deleted[/]")


@managercli.command()
def shell():
    """starts a IPython REPL console with db objects and models"""
    from IPython import start_ipython

    db = SQL(settings.SQL)
    Session = db.sessionmaker()
    start_ipython(
        argv=[],
        user_ns={
            "settings": settings,
            "Session": Session,
            "db": db,
            "session": Session(),
        },
    )


# managercli.add_command(db)
# managercli.add_command(users)



import logging
from contextlib import asynccontextmanager

from sqlalchemy import MetaData
from sqlalchemy.ext.asyncio import AsyncSession, create_async_engine

# from sqlalchemy.dialects.postgresql import ARRAY
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql.selectable import Select

from labfunctions.db.common import Base


class AsyncSQL:
    """
    As example checks:
    https://docs.sqlalchemy.org/en/14/_modules/examples/asyncio/async_orm.html
    """

    Meta = MetaData()

    def __init__(self, uri, echo=False):
        self._session = None
        self._engine = None
        self._uri = uri
        self._echo = echo
        # self.meta = MetaData()

    def __getattr__(self, name):
        return getattr(self._session, name)

    @property
    def engine(self):
        return self._engine

    async def init(self, pool_size=20, max_overflow=0):
        if "sqlite" in self._uri.split("://", maxsplit=1)[0]:
            self._engine = create_async_engine(self._uri)
        else:
            self._engine = create_async_engine(
                self._uri,
                echo=self._echo,
                pool_size=pool_size,
                max_overflow=max_overflow,
            )
        self.Meta.bind = self._engine

    def sessionmaker(self, expire_on_commit=False):
        """
        expire_on_commit=False will prevent attributes from being expired
        after commit
        """
        return sessionmaker(
            self._engine, expire_on_commit=expire_on_commit, class_=AsyncSession
        )()

    async def session_execute(self, stmt: Select):
        _session = self.sessionmaker()
        async with _session.begin():
            result = await _session.execute(stmt)
            return result

    @asynccontextmanager
    async def session(self):
        async_session = sessionmaker(
            self._engine, expire_on_commit=False, class_=AsyncSession
        )
        async with async_session() as session:
            async with session.begin():
                yield session

    async def create_all(self):
        async with self._engine.begin() as conn:
            await conn.run_sync(Base.metadata.create_all)

    async def drop_all(self):
        async with self._engine.begin() as conn:
            await conn.run_sync(Base.metadata.drop_all)



from datetime import datetime
from typing import Any, Dict, Optional

from pydantic import BaseModel


class DockerBuildLowLog(BaseModel):
    logs: str
    error: bool


class DockerPushLog(BaseModel):
    logs: str
    error: bool


class DockerBuildLog(BaseModel):
    build_log: DockerBuildLowLog
    push_log: Optional[DockerPushLog] = None
    error: bool


class DockerResources(BaseModel):
    mem_limit: Optional[int] = None
    mem_reservation: Optional[int] = None


class DockerVolume(BaseModel):
    orig_mount: str
    dst_mount: str
    extra: Dict[str, Any] = {}


class DockerRunResult(BaseModel):
    msg: str
    status: int


class DockerfileImage(BaseModel):
    maintener: str
    image: str
    user: Dict[str, int] = {"uid": 1089, "gid": 1090}
    build_packages: str = "build-essential libopenblas-dev git"
    final_packages: Optional[str] = None


class RuntimeVersionOrm(BaseModel):
    docker_name: str
    version: str
    project_id: str
    created_at: datetime
    id: Optional[int] = None

    class Config:
        orm_mode = True


class RuntimeVersionData(BaseModel):
    docker_name: str
    version: str
    projectid: str
    id: Optional[int] = None



# pylint: disable=too-few-public-methods
from datetime import datetime

from sqlalchemy import (
    BigInteger,
    Boolean,
    Column,
    DateTime,
    Float,
    ForeignKey,
    Integer,
    String,
    UniqueConstraint,
)

# from sqlalchemy.dialects.postgresql import BYTEA
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.orm import declarative_mixin, declared_attr, relationship
from sqlalchemy.schema import Table
from sqlalchemy.sql import functions
from sqlalchemy.types import BINARY, JSON
from sqlalchemy_serializer import SerializerMixin

from labfunctions.db.common import Base

assoc_projects_users = Table(
    "nb_projects_users",
    Base.metadata,
    Column("project_id", ForeignKey("nb_project.projectid")),
    Column("user_id", ForeignKey("nb_user.id")),
)


@declarative_mixin
class ProjectRelationMixin:
    # pylint: disable=no-self-argument

    @declared_attr
    def project_id(cls):
        return Column(
            String(16),
            ForeignKey("nb_project.projectid", ondelete="SET NULL"),
            nullable=True,
        )

    @declared_attr
    def project(cls):
        return relationship("ProjectModel")


@declarative_mixin
class UserMixin:

    id = Column(BigInteger, primary_key=True)
    password = Column(BINARY, nullable=True)
    is_superuser = Column(Boolean, default=False, nullable=False)
    is_active = Column(Boolean, default=True, nullable=False)
    scopes = Column(String(), default="user", nullable=False)

    created_at = Column(
        DateTime(),
        server_default=functions.now(),
        nullable=False,
    )
    updated_at = Column(
        DateTime(),
        server_default=functions.now(),
        nullable=False,
    )


class HistoryModel(Base, SerializerMixin, ProjectRelationMixin):
    """
    Register each execution of a workflow.

    :param wfid: Is the workflow wfid
    :param taskid: is random id generated for each execution of the workflow
    :param name: the filename of the notebook executed
    :param result: is the result of the task. TaskResult
    :param elapsed_secs: Time in seconds from the start of the task to the end.
    :param status: -1 fail, 0 ok.
    """

    __tablename__ = "nb_history"
    __mapper_args__ = {"eager_defaults": True}

    id = Column(BigInteger, primary_key=True)
    wfid = Column(String(24))
    execid = Column(String(24))  # should be execution id
    nb_name = Column(String(), nullable=False)
    result = Column(JSON, nullable=False)
    elapsed_secs = Column(Float(), nullable=False)
    status = Column(Integer, index=True)

    created_at = Column(
        DateTime(),
        server_default=functions.now(),
        index=True,
        nullable=False,
    )


class ProjectModel(Base, SerializerMixin):
    """
    Register each execution of a workflow.

    :param wfid: Is the workflow wfid
    :param taskid: is random id generated for each execution of the workflow
    :param name: the filename of the notebook executed
    :param result: is the result of the task. TaskResult
    :param elapsed_secs: Time in seconds from the start of the task to the end.
    :param status: -1 fail, 0 ok.
    """

    __tablename__ = "nb_project"
    __mapper_args__ = {"eager_defaults": True}

    id = Column(BigInteger, primary_key=True)
    projectid = Column(String(16), index=True, unique=True, nullable=False)
    name = Column(String(128), unique=True, nullable=False)
    private_key = Column(BINARY, nullable=False)
    description = Column(String())
    repository = Column(String(2048), nullable=True)
    owner_id = Column(
        BigInteger,
        ForeignKey("nb_user.id", ondelete="SET NULL"),
        nullable=True,
    )
    owner = relationship(
        "labfunctions.models.UserModel", foreign_keys="ProjectModel.owner_id"
    )
    # agent_id = Column(
    #     BigInteger,
    #     ForeignKey("nb_user.id", ondelete="SET NULL"),
    #     nullable=True,
    # )
    # agent = relationship(
    #     "labfunctions.models.UserModel", foreign_keys="ProjectModel.agent_id"
    # )
    users = relationship(
        "labfunctions.models.UserModel",
        secondary=assoc_projects_users,
        back_populates="projects",
    )
    created_at = Column(
        DateTime(),
        server_default=functions.now(),
        nullable=False,
    )
    updated_at = Column(
        DateTime(),
        server_default=functions.now(),
        nullable=False,
    )


class UserModel(UserMixin, Base):
    __tablename__ = "nb_user"
    __mapper_args__ = {"eager_defaults": True}

    username = Column(String(), index=True, unique=True, nullable=False)
    email = Column(String(), index=True, nullable=True)
    is_agent = Column(Boolean(), index=True, nullable=False, default=False)
    projects = relationship(
        "ProjectModel", secondary=assoc_projects_users, back_populates="users"
    )


class WorkflowModel(Base, SerializerMixin, ProjectRelationMixin):
    """
    Configuration for each workflow.

    :param wfid: an unique identifier for this workflow
    :param alias: because the filename could be shared between different
    workflows, an alias was added to identify each instance, and is more
    friendly than wfid.
    :param name: name of the notebook file.
    :param description: A friendly description of the purpose of this workflow
    :param nbtask: details to execute a notebook with specific parameters.
    :param schedule: when should be executed.
    :param enabled: if the task should run or not.
    """

    __tablename__ = "nb_workflow"
    __table_args__ = (
        UniqueConstraint("alias", "project_id", name="_nb_workflow__project_alias"),
    )
    # needed for async support
    __mapper_args__ = {"eager_defaults": True}

    id = Column(Integer, primary_key=True)
    wfid = Column(String(24), index=True, unique=True)
    alias = Column(String(33), index=True, nullable=False)
    # nb_name = Column(String(), nullable=False)
    nbtask = Column(JSON(), nullable=False)
    schedule = Column(JSON(), nullable=False)
    enabled = Column(Boolean, default=True, nullable=False)

    created_at = Column(DateTime(), server_default=functions.now(), nullable=False)

    updated_at = Column(DateTime(), server_default=functions.now())


class NotebookFile(Base, ProjectRelationMixin):
    """
    Notebook Register
    """

    __tablename__ = "nb_notebook_file"
    __table_args__ = (
        UniqueConstraint(
            "nb_name", "project_id", name="_nb_notebook_file__name_project"
        ),
    )
    __mapper_args__ = {"eager_defaults": True}
    id = Column(Integer, primary_key=True)
    nb_name = Column(String(24), index=True)
    remote_path = Column(String(24), nullable=True)

    params = Column(String(), nullable=True)
    runtimeid = Column(String(), index=True, nullable=False)
    owner = relationship(
        "labfunctions.models.UserModel", foreign_keys="NotebookFile.owner_id"
    )
    owner_id = Column(
        BigInteger,
        ForeignKey("nb_user.id", ondelete="SET NULL"),
        nullable=False,
    )
    created_at = Column(DateTime(), server_default=functions.now(), nullable=False)


class RuntimeModel(Base, ProjectRelationMixin):
    """
    Runtimes Register
    runtimeid = [projectid]/[runtime_name]/[version]
    runtime_name = should be nbworkflows/[projectid]-[runtime_name]
    """

    __tablename__ = "nb_runtime"
    __mapper_args__ = {"eager_defaults": True}

    id = Column(Integer, primary_key=True)
    runtimeid = Column(String(), unique=True, index=True, nullable=False)
    runtime_name = Column(String(), index=True, nullable=False)
    docker_name = Column(String(), nullable=False)
    spec = Column(JSON(), nullable=False)
    version = Column(String(), nullable=False)
    registry = Column(String(), nullable=True)

    created_at = Column(
        DateTime(), server_default=functions.now(), nullable=False, index=True
    )


class MachineModel(Base):

    __tablename__ = "nb_machine"
    __mapper_args__ = {"eager_defaults": True}
    __table_args__ = (
        UniqueConstraint("name", "provider", name="_nb_machine__name_provider"),
    )

    id = Column(Integer, primary_key=True)
    name = Column(String(), unique=True, index=True, nullable=False)
    location = Column(String(), unique=False, index=True, nullable=False)
    provider = Column(String(), nullable=False, index=True)
    desc = Column(String(), nullable=True)
    machine_type = Column(JSON(), nullable=False)
    gpu = Column(JSON(), nullable=True)
    volumes = Column(JSON(), nullable=True)

    created_at = Column(
        DateTime(), server_default=functions.now(), nullable=False, index=True
    )
    updated_at = Column(
        DateTime(),
        server_default=functions.now(),
        nullable=False,
    )



from typing import Optional, Union

import aioredis
from aioredis.client import Redis

from .base import TokenStoreSpec
from .utils import generate_token


class RedisTokenStore(TokenStoreSpec):
    def __init__(self, redis: Union[Redis, str], namespace: str = "rtkn"):
        if isinstance(redis, str):
            _redis = aioredis.from_url(redis, decode_responses=True)
        else:
            _redis = redis
        self.redis: Redis = _redis
        self.ns = namespace

    async def put(self, key: str, value: str, ttl: Optional[int] = None) -> bool:
        await self.redis.set(f"{self.ns}:{key}", value)
        return True

    async def get(self, key: str) -> str:
        val = await self.redis.get(f"{self.ns}:{key}")
        value = val
        if not isinstance(val, str) and val:
            value = val.decode("utf-8")

        return value

    async def delete(self, key: str):
        await self.redis.delete(f"{self.ns}:{key}")

    @staticmethod
    def generate(sign: Optional[str] = None) -> str:
        return generate_token()



import nbformat
from nbconvert import NotebookExporter
from traitlets.config import Config


def read_notebook(path, version=4):
    with open(path, "r", encoding="utf-8") as f:
        data = f.read()
        note = nbformat.reads(data, as_version=version)
    return note


def clean_output(path, version=4):
    """
    Cleans notebooks outputs, for more info see:
    https://nbconvert.readthedocs.io/en/latest/api/preprocessors.html#nbconvert.preprocessors.ClearOutputPreprocessor
    https://nbconvert.readthedocs.io/en/latest/nbconvert_library.html
    """
    c = Config()
    c.NotebookExporter.preprocessors = [
        "nbconvert.preprocessors.ClearOutputPreprocessor"
    ]

    exporter = NotebookExporter(config=c)

    note = read_notebook(path, version=version)
    exported, resources_dict = exporter.from_notebook_node(note)

    return exported, resources_dict



import json
import logging
import os
from pathlib import Path
from typing import Any, Callable, Dict, Generator, List, Optional, Union

import httpx
import jwt

from labfunctions import defaults, errors, types
from labfunctions.events import EventManager
from labfunctions.hashes import generate_random
from labfunctions.utils import mkdir_p, open_yaml, write_yaml

from .state import WorkflowsState
from .utils import get_credentials_disk


def get_http_client(**kwargs) -> httpx.Client:

    return httpx.Client(**kwargs)


class AuthFlow(httpx.Auth):
    requires_request_body = True
    requires_response_body = True

    def __init__(self, access_token, refresh_token, refresh_url):
        self.access_token = access_token
        self.refresh_token = refresh_token
        self.refresh_url = refresh_url

    def auth_flow(self, request):
        request.headers["Authorization"] = f"Bearer {self.access_token}"
        response = yield request

        # is_valid = validate_credentials_local(self.access_token)
        if response.status_code == 401:
            # If the server issues a 401 response, then issue a request to
            # refresh tokens, and resend the request.
            refresh_response = yield self.build_refresh_request()
            self.update_tokens(refresh_response)

            request.headers["Authorization"] = f"Bearer {self.access_token}"
            yield request

    def build_refresh_request(self):
        # Return an `httpx.Request` for refreshing tokens.
        rtkn = {"refresh_token": self.refresh_token}
        req = httpx.Request("POST", self.refresh_url, json=rtkn)
        req.headers["Authorization"] = f"Bearer {self.access_token}"
        return req

    def update_tokens(self, response):
        # Update the `.access_token` and `.refresh_token` tokens
        # based on a refresh response.
        data = response.json()
        tkn = data.get("access_token")
        if not tkn:
            raise errors.AuthValidationFailed()
        self.access_token = tkn


class BaseClient:
    """
    A Generic client for API Server communication.


    By default, each client is associated to a specific project,
    this could be decoupled in the future because some actions
    like **startproject** are independent of the project.

    **Some conventions:**

    If a method calls to an API's endpoint, it should use the name
    of the endpoint at first. For instance:

        GET /workflows/<projectid>/<wfid>
    The method name for that endpoint could be: `workflows_get_one()`


    :param url_service: base url of the WORKFLOWS_SERVICE
    :param projectid: projectid realted to ProjectsModel
    :param creds: TokenCreds type, it includes access_token and refresh_token
    :param store_creds: Optional, if true the credentials will be stored on disk
    :param project: Optional[Project] type
    :param version: version api, not implemented yet
    :param http_init_func: Callable, a function which initializes
    a HTTPX client.
    """

    def __init__(
        self,
        url_service: str,
        creds: Optional[types.TokenCreds] = None,
        wf_state: Optional[WorkflowsState] = None,
        version=defaults.API_VERSION,
        http_init_func=get_http_client,
        timeout=defaults.CLIENT_TIMEOUT,
        base_path: Optional[str] = None,
    ):
        self._addr = url_service
        self._creds = creds
        self._auth: Optional[AuthFlow] = None
        self.logger = logging.getLogger("nbwork.agent")
        self.state = wf_state
        self._version = version
        self._timeout = timeout
        self._http_creator = http_init_func
        self._http: httpx.Client = self._http_init()
        self.logger = logging.getLogger("nbwork.client")
        self.base_path = base_path
        self._home = Path.home() / defaults.CLIENT_HOME_DIR

    @property
    def working_area(self) -> Path:
        """Working area is used to store related project data as private keys
        temp build generation files and so on"""

        return self._home / self.projectid

    @property
    def homedir(self) -> Path:
        return self._home

    def create_homedir(self) -> Path:
        if not self._home.is_dir():
            mkdir_p(self._home)
            self._home.chmod(0o700)
        return self._home

    def info(self) -> Dict[str, Any]:
        workflows = None
        if self.state.workflows:
            workflows = list(self.state.workflows.keys())

        return {
            "projectid": self.projectid,
            "project_name": self.project_name,
            "workflows_service": self._addr,
            "homedir": str(self.homedir),
            "working_area": str(self.working_area),
            "file": self.state.workflows_file,
            "workflows": workflows,
            "user": self.user,
            "base_path": self.base_path,
        }

    @property
    def http(self) -> httpx.Client:
        return self._http

    @property
    def projectid(self) -> str:
        if self.state:
            return self.state.projectid
        else:
            raise errors.client.WorkflowStateNotSetError(__name__)

    @property
    def project_name(self) -> str:
        if self.state:
            return self.state.project_name
        else:
            raise errors.client.WorkflowStateNotSetError(__name__)

    # def get_context(self, execid=None) -> ExecutionNBTask:
    #    _env = os.getenv(defaults.EXECUTIONTASK_VAR)
    #    if _env:
    #        ctx = ExecutionNBTask(**json.loads(_env))
    #    else:
    #        ctx = context.create_dummy_ctx(self.projectid, execid)
    #    return ctx

    def _auth_init(self) -> AuthFlow:
        self._auth = AuthFlow(
            self._creds.access_token,
            self._creds.refresh_token,
            f"{self._addr}/{self._version}/{defaults.REFRESH_TOKEN_PATH}",
        )

    def _http_init(self) -> httpx.Client:
        """When token is updated the client MUST BE updated too."""
        # _headers = {"Authorization": f"Bearer {self.creds.access_token}"}

        if self._creds:
            self._auth_init()
        return self._http_creator(
            base_url=f"{self._addr}/{self._version}",
            timeout=self._timeout,
            auth=self._auth,
        )

    @property
    def user(self) -> Union[str, None]:
        if self._creds:
            decoded = jwt.decode(
                self._creds.access_token, options={"verify_signature": False}
            )

            return decoded["usr"]
        return None

    @property
    def creds(self) -> types.TokenCreds:
        return types.TokenCreds(
            access_token=self._auth.access_token,
            refresh_token=self._auth.refresh_token,
        )

    @creds.setter
    def creds(self, creds: types.TokenCreds):
        """
        If credentials are set, the http client should be
        re-initialized
        """
        self._creds = creds
        self._http = self._http_init()

    def load_creds(self):
        at = os.getenv("NB_AGENT_TOKEN")
        rt = os.getenv("NB_AGENT_REFRESH_TOKEN")
        if at and rt:
            self.creds = types.TokenCreds(access_token=at, refresh_token=rt)
        self.creds = get_credentials_disk(self.homedir)

    def login(self, u: str, p: str):
        rsp = httpx.post(
            f"{self._addr}/{self._version}/auth/login",
            json=dict(username=u, password=p),
            timeout=self._timeout,
        )
        if rsp.status_code == 200:
            self.creds = types.TokenCreds(**rsp.json())
            # self._http = self.http_init()
        else:
            raise errors.client.LoginError(self._addr, u)

    def verify(self):
        try:
            rsp = self._http.get(f"/auth/verify")
            if rsp.status_code == 200:
                return True
        except errors.AuthValidationFailed:
            self._creds = None
            self._auth = None
            return False

    def write(self, output="workflows.yaml"):
        self.state.write(output)

    def close(self):
        self._http.close()

    def events_listen(
        self, execid, last=None, timeout=None
    ) -> Generator[types.events.EventSSE, None, None]:
        timeout = timeout or self._timeout
        # uri = f"/events/{self.projectid}/{execid}/_listen"
        uri = f"{self._addr}/{self._version}/events/{self.projectid}/{execid}/_listen"
        if last:
            uri = f"{uri}?last={last}"

        headers = {"Authorization": f"Bearer {self.creds.access_token}"}

        with httpx.stream("GET", uri, timeout=timeout, headers=headers) as r:
            buffer_ = ""
            for line in r.iter_lines():
                buffer_ += line
                if buffer_.endswith("\n\n"):
                    evt = EventManager.from_sse2event(buffer_)
                    # print(evt.dict())
                    if evt.data == "exit":
                        return
                    yield evt
                    buffer_ = ""

    def events_publish(self, execid, data, event=None):
        final = data
        if isinstance(data, dict):
            final = json.dumps(data)
        if final:
            evt = types.events.EventSSE(data=final, event=event)
            self._http.post(
                f"/events/{self.projectid}/{execid}/_publish", json=evt.dict()
            )
        else:
            self.logger.warning(f"execid: {execid} empty message")



import os

import click
from rich.console import Console

from labfunctions.cmd.common import info, login, startproject


def load_client_cli(cli):
    from labfunctions.cmd.executors import executorscli
    from labfunctions.cmd.history import statuscli
    from labfunctions.cmd.project import projectcli
    from labfunctions.cmd.runtimes import runtimescli
    from labfunctions.cmd.workflows import workflowscli

    cli.add_command(workflowscli)
    cli.add_command(projectcli)
    cli.add_command(statuscli)
    cli.add_command(executorscli)
    cli.add_command(runtimescli)


def load_server_cli(cli):
    from labfunctions.cmd.agent import agentcli
    from labfunctions.cmd.cluster import clustercli
    from labfunctions.cmd.manager import managercli
    from labfunctions.cmd.runtimes import runtimescli
    from labfunctions.cmd.services import schedulercli, webcli

    cli.add_command(managercli)
    cli.add_command(webcli)
    cli.add_command(schedulercli)
    cli.add_command(agentcli)
    cli.add_command(clustercli)
    cli.add_command(runtimescli)


def init_cli():

    console = Console()

    @click.group()
    @click.pass_context
    def cli(ctx):
        """
        NB Workflow command line tool
        """

    @click.command()
    @click.pass_context
    def version(ctx):
        """Actual version"""
        from labfunctions.utils import get_version

        ver = get_version("__version__.py")
        console.print(f"[bold magenta]{ver}[/bold magenta]")

    if os.environ.get("NB_SERVER", False):
        load_server_cli(cli)

        if os.environ.get("DEBUG", False):
            # cli.add_command(executorscli)
            load_client_cli(cli)
    else:
        load_client_cli(cli)

    cli.add_command(startproject)
    cli.add_command(login)
    cli.add_command(version)
    cli.add_command(info)
    return cli


cli = init_cli()

if __name__ == "__main__":

    cli(ctx={})
    # cli(obj={})



from typing import List, Optional

from pydantic import BaseModel

from .security import JWTResponse


class UserOrm(BaseModel):
    username: str
    id: Optional[int] = None
    email: Optional[str] = None
    password: Optional[bytes] = None
    scopes: str = "user:r:w"
    is_superuser: bool = False
    is_active: bool = False
    is_agent: bool = False
    projects: List[str] = []

    class Config:
        orm_mode = True


class AgentReq(BaseModel):
    agent_name: str


class AgentJWTResponse(BaseModel):
    agent_name: str
    creds: JWTResponse



# pylint: disable=unused-argument
import json as std_json
import pathlib
from typing import List, Union

import httpx
from sanic import Blueprint, Request, Sanic, exceptions
from sanic.response import empty, json
from sanic_ext import openapi

from labfunctions import defaults, types
from labfunctions.conf.server_settings import settings
from labfunctions.defaults import API_VERSION
from labfunctions.managers import projects_mg, users_mg
from labfunctions.runtimes.context import build_upload_uri, create_build_ctx
from labfunctions.security import get_auth
from labfunctions.security.web import protected
from labfunctions.utils import run_async, secure_filename
from labfunctions.web.utils import (
    get_kvstore,
    get_query_param2,
    get_scheduler,
    stream_reader,
)

projects_bp = Blueprint("projects", url_prefix="projects", version=API_VERSION)


def get_token_data(request: Request):
    return request.ctx.token_data


@projects_bp.post("/")
@openapi.body({"application/json": types.ProjectReq})
@openapi.response(201, types.ProjectData, "Created")
@protected()
@users_mg.inject_user
async def project_create(request, user: types.user.UserOrm):
    """Create a new project"""
    # pylint: disable=unused-argument

    dict_ = request.json
    token_data = get_token_data(request)
    pr = types.ProjectReq(**dict_)
    session = request.ctx.session
    r = await projects_mg.create(session, user.id, pr)
    if r:
        return json(r.dict(), 201)
    return json(dict(msg="already exist"), 200)


@projects_bp.put("/")
@openapi.body({"application/json": types.ProjectReq})
@openapi.response(202, "created")
@protected()
@users_mg.inject_user
async def project_create_or_update(request, user: types.user.UserOrm):
    """Create or update a project"""
    # pylint: disable=unused-argument

    dict_ = request.json
    pd = types.ProjectReq(**dict_)
    session = request.ctx.session
    async with session() as new_session:
        r = await projects_mg.create_or_update(new_session, user.id, pd)
    return json(dict(msg="created"), 202)


@projects_bp.get("/")
@openapi.response(200, List[types.ProjectData], "project-list")
@protected()
@users_mg.inject_user
async def project_list(request, user: types.user.UserOrm):
    """Get a list of projects belonging to a user"""
    # pylint: disable=unused-argument

    session = request.ctx.session
    result = await projects_mg.list_all(session, user.id)
    return json([r.dict() for r in result], 200)


@projects_bp.get("/<projectid:str>")
@openapi.parameter("projectid", str, "path")
@openapi.response(200, "project")
@openapi.response(404, "not found")
@protected()
@users_mg.inject_user
async def project_get_one(request, projectid, user: types.user.UserOrm):
    """Get one project"""
    # pylint: disable=unused-argument

    session = request.ctx.session
    print("User id: ", user.id)
    r = await projects_mg.get_by_projectid(session, projectid, user_id=user.id)
    if r:
        return json(r.dict(), 200)
    return json(dict(msg="Not found"))


@projects_bp.delete("/<projectid:str>")
@openapi.parameter("projectid", str, "path")
@openapi.response(200, "project")
@openapi.response(404, "not found")
@protected()
async def project_delete(request, projectid):
    """Delete Project"""
    # pylint: disable=unused-argument

    session = request.ctx.session
    await projects_mg.delete_by_projectid(session, projectid)
    return json(dict(msg="deleted"))


@projects_bp.post("/<projectid:str>/agent")
@openapi.parameter("projectid", str, "path")
@protected()
async def project_create_user_agent(request, projectid):
    """
    Creates a User Agent
    """
    session = request.ctx.session
    async with session.begin():
        res = await projects_mg.create_agent(session, projectid)
        if res:
            return json(dict(msg=res.username), 201)

    return json(dict(msg="not created"), 200)


@projects_bp.delete("/<projectid:str>/agent/<agentname:str>")
@openapi.parameter("projectid", str, "path")
@openapi.parameter("agentname", str, "path")
@protected()
async def project_delete_user_agent(request, projectid, agentname):
    """
    Delete a User Agent
    """
    session = request.ctx.session
    await projects_mg.delete_agent(session, agentname, projectid)
    await session.commit()
    return json(dict(msg="deleted"), 200)


@projects_bp.post("/<projectid:str>/agent/_token")
@openapi.parameter("projectid", str, "path")
@openapi.response(200, types.user.AgentJWTResponse, "agent credentials")
@protected()
async def project_get_last_agent_token(request, projectid: str):
    """
    Get the agent token by a project
    """
    # pylint: disable=unused-argument
    _auth = get_auth(request)
    session = request.ctx.session
    # redis = request.ctx.web_redis

    agt = await projects_mg.get_agent_project(session, projectid)
    if agt:
        jwt = await users_mg.get_jwt_token(_auth, agt, exp=settings.AGENT_TOKEN_EXP)
        agt_jwt = types.user.AgentJWTResponse(agent_name=agt.username, creds=jwt)
        return json(agt_jwt.dict(), 200)
    return json({"msg": "agent not found"}, 404)


@projects_bp.post("/<projectid:str>/agent/<agent:str>/_token")
@openapi.parameter("projectid", str, "path")
@openapi.parameter("agent", str, "path")
@openapi.response(200, types.user.AgentJWTResponse, "agent credentials")
@protected()
async def project_get_agent_token(request, projectid: str, agent: str):
    """
    Get the agent token by a project
    """
    # pylint: disable=unused-argument
    _auth = get_auth(request)
    session = request.ctx.session
    redis = request.ctx.web_redis

    agt = await projects_mg.get_agent(session, agent, projectid)
    jwt = await users_mg.get_jwt_token(_auth, agt, exp=settings.AGENT_TOKEN_EXP)
    agt_jwt = types.user.AgentJWTResponse(agent_name=agt.username, creds=jwt)

    return json(agt_jwt.dict(), 200)


@projects_bp.get("/<projectid:str>/agent")
@openapi.parameter("projectid", str, "path")
@openapi.response(200, List[str], "agent credentials")
@protected()
async def project_agents_list(request, projectid: str):
    """
    Get the agent token by a project
    """
    # pylint: disable=unused-argument
    session = request.ctx.session

    agents = await projects_mg.get_agent_list(session, projectid)

    return json(agents, 200)


@projects_bp.post("/<projectid:str>/_build")
@openapi.body({"application/json": types.RuntimeSpec})
@openapi.parameter("version", str, "query")
@protected()
async def project_build(request, projectid):
    """
    Enqueue docker build image
    """
    # pylint: disable=unused-argument
    spec = types.RuntimeSpec(**request.json)

    root = pathlib.Path(projectid)

    version = get_query_param2(request, "version", None)

    # session = request.ctx.session
    # async with session.begin():
    #    pd = await projects_mg.get_by_projectid(session, projectid)
    # if pd:
    ctx = create_build_ctx(
        projectid,
        spec,
        version,
        project_store_class=settings.PROJECTS_STORE_CLASS_SYNC,
        project_store_bucket=settings.PROJECTS_STORE_BUCKET,
        registry=settings.DOCKER_REGISTRY,
    )

    sche = get_scheduler(request, settings.BUILD_QUEUE)
    job = await run_async(sche.enqueue_build, ctx)

    return json(dict(msg="ok", execid=job.id), 202)
    # return json(dict(msg="not found"), 404)


@projects_bp.post("/<projectid:str>/_upload", stream=True)
@openapi.parameter("version", str, "query")
@openapi.parameter("runtime", str, "query")
@protected()
async def project_upload(request, projectid):
    """
    Upload a workflow project
    """
    # pylint: disable=unused-argument

    root = pathlib.Path(projectid)
    version = get_query_param2(request, "version", None)
    runtime_name = get_query_param2(request, "runtime", None)
    session = request.ctx.session
    async with session.begin():
        pd = await projects_mg.get_by_projectid(session, projectid)

    uri = build_upload_uri(pd.projectid, runtime_name, version)

    kv = get_kvstore(request)
    rsp = await kv.put_stream(uri, stream_reader(request))
    if rsp:
        return json(dict(msg="ok"), 201)
    else:
        return empty()

    # fileserver = f"{settings.FILESERVER}/{settings.FILESERVER_BUCKET}"
    # dst_url = f"{fileserver}/{uri}"
    # async with httpx.AsyncClient() as client:
    #    r = await client.put(dst_url, content=stream_reader(request))
    # if r.status_code == 204:
    #    return empty()


@projects_bp.get("/<projectid:str>/_private_key")
@openapi.parameter("projectid", str, "path")
@openapi.response(200, "project")
@openapi.response(404, "not found")
@protected()
async def project_private_key(request, projectid):
    """Get private key based on the project"""
    # pylint: disable=unused-argument

    session = request.ctx.session
    async with session.begin():
        r = await projects_mg.get_private_key(session, projectid)
        if r:
            return json({"private_key": r}, 200)

    return json(dict(msg="Not found"))



from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field

from labfunctions import defaults

from .docker import DockerfileImage


class ScheduleData(BaseModel):
    """Used as generic structure when querying database"""

    start_in_min: int = 0
    repeat: Optional[int] = None
    cron: Optional[str] = None
    interval: Optional[str] = None


class NBTask(BaseModel):
    """
    NBTask is the task definition. It will be executed by papermill.
    This interface is used together with the ScheduleCron or ScheduleInterval
    to define a job.

    :param nb_name: is the name of the notebook to run
    :param params: a dict with the params to run the specific notebook,
    wrapper around papermill.
    :param remote_input: optional to get a remote notebook (not implemented)
    :param remote_output: optional to put in a remote place (not implemented)
    :param runtime: name of the runtime (`runtime_name`)
    :param version: Version of the runtime
    :param machine: where to run
    :param gpu_support: if it needs gpu support
    :param timeout: time in secs to wait from the start of the task
    to mark the task as failed.
    :param notifications_ok: If ok send a notification to discord or slack.
    :param notifications_fail: If not ok, send notification to discord or slack.
    but internally the task also send a notification if the user wants.
    """

    nb_name: str
    params: Dict[str, Any]
    remote_input: Optional[str] = None
    remote_output: Optional[str] = None
    runtime: Optional[str] = None
    version: Optional[str] = None
    machine: str = defaults.MACHINE_TYPE
    cluster: str = defaults.CLUSTER_NAME
    gpu_support: bool = False
    description: Optional[str] = None
    timeout: int = 10800  # secs 3h default
    notifications_ok: Optional[List[str]] = None
    notifications_fail: Optional[List[str]] = None
    # schedule: Optional[ScheduleData] = None


class ExecutionNBTask(BaseModel):
    """It will be send to task_handler, and it has the
    configuration needed for papermill to run a specific notebook.
    """

    projectid: str
    wfid: str
    execid: str
    nb_name: str
    params: Dict[str, Any]
    runtime: str
    # folders:
    pm_input: str
    pm_output: str
    output_name: str

    output_dir: str
    error_dir: str

    today: str
    timeout: int
    created_at: str
    gpu_support: bool = False
    cluster: str = defaults.CLUSTER_NAME
    machine: str = defaults.MACHINE_TYPE
    remote_input: Optional[str]
    remote_output: Optional[str]
    notifications_ok: Optional[List[str]] = None
    notifications_fail: Optional[List[str]] = None


class ExecutionResult(BaseModel):
    """
    Is the result of a ExecutionTask execution.
    """

    projectid: str
    execid: str
    wfid: str

    name: str
    params: Dict[str, Any]
    input_: str
    error: bool
    elapsed_secs: float
    created_at: str
    cluster: Optional[str] = None
    machine: Optional[str] = None
    runtime: Optional[str] = None
    output_name: Optional[str] = None
    output_dir: Optional[str] = None
    error_dir: Optional[str] = None
    error_msg: Optional[str] = None


@dataclass
class SimpleExecCtx:
    wfid: str
    execid: str
    execution_dt: str


class HistoryResult(BaseModel):
    wfid: str
    # posible status: queued, started, deferred,
    # finished, stopped, scheduled, canceled, failed.
    status: int
    result: Optional[ExecutionResult] = None
    execid: Optional[str] = None
    created_at: Optional[str] = None


class HistoryRequest(BaseModel):
    task: NBTask
    result: ExecutionResult


class HistoryLastResponse(BaseModel):
    rows: List[HistoryResult]


class WorkflowData(BaseModel):
    wfid: str
    alias: str
    nbtask: Dict[str, Any]
    enabled: bool = True
    schedule: Optional[ScheduleData] = None


class WorkflowDataWeb(BaseModel):
    alias: str
    nbtask: NBTask
    enabled: bool = True
    wfid: Optional[str] = None
    schedule: Optional[ScheduleData] = None


@dataclass
class WorkflowsList:
    rows: List[WorkflowData]



import importlib
import json
import logging
import logging.config
import os
import subprocess
import sys
from pathlib import Path

from labfunctions import defaults
from labfunctions.types import ClientSettings, ServerSettings

# from labfunctions.types.config import SecuritySettings

# from labfunctions.utils import define_base_path
# from logging import NullHandler

# Server defaults
GLOBAL_MODULE = "labfunctions.conf.global_settings"
ENVIRONMENT_VARIABLE = "NB_SETTINGS_MODULE"
DEFAULT_MODULE = os.environ.get(ENVIRONMENT_VARIABLE, GLOBAL_MODULE)

# Client defaults
GLOBAL_CLIENT = "labfunctions.conf.global_client"
CLIENT_VARIABLE = "NB_CLIENT_MODULE"
DEFAULT_CLIENT_MOD = os.environ.get(CLIENT_VARIABLE, "nb_app.settings")


def _get_level(level):
    return getattr(logging, level)


def execute_cmd(cmd) -> str:
    """Wrapper around subprocess"""
    with subprocess.Popen(
        cmd.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE
    ) as p:

        out, err = p.communicate()
        if err:
            raise AttributeError(err.decode())

        return out.decode().strip()


def define_base_path() -> str:
    """
    It tries to determine a base path for BASE_PATH in settings.
    If a environment var exist for this, then it will use that, else
    it will use git, if fail it will go to an upper level.

    """
    root_dir = os.getcwd()
    base_path = root_dir

    base_var = os.environ.get(defaults.BASE_PATH_ENV)
    # is_interactive = _is_interactive_shell()
    if not base_var:
        try:
            base_path = execute_cmd("git rev-parse --show-toplevel")
        except:
            base_path = str(Path(root_dir).parents[0])
    elif base_var:
        base_path = base_var

    return base_path


def load_client(settings_module=DEFAULT_CLIENT_MOD) -> ClientSettings:
    # sys.path.append(os.getcwd())
    module_loaded = settings_module
    base_path = define_base_path()

    if not os.environ.get(CLIENT_VARIABLE):
        sys.path.append(base_path)

    try:
        mod = importlib.import_module(settings_module)
    except ModuleNotFoundError:
        mod = importlib.import_module(GLOBAL_CLIENT)
        module_loaded = GLOBAL_CLIENT

    settings_dict = {}
    for m in dir(mod):
        if m.isupper():
            # sets.add(m)
            value = getattr(mod, m)
            settings_dict[m] = value

    bp = settings_dict.get("BASE_PATH")
    if bp:
        base_path = bp
    else:
        settings_dict["BASE_PATH"] = base_path

    cfg = ClientSettings(**settings_dict)
    cfg.SETTINGS_MODULE = module_loaded
    if not cfg.DEBUG:
        _level = _get_level(cfg.LOGLEVEL)
    else:
        _level = logging.DEBUG

    # set BASE_PATH
    # os.environ[defaults.BASE_PATH_ENV] = cfg.BASE_PATH

    # logging.basicConfig(format=cfg.LOGFORMAT, level=_level)
    logging.config.dictConfig(cfg.LOGCONFIG)
    log = logging.getLogger("nbwork.client")
    log.debug("Using {cfg.SETTINGS_MODULE} as config module")

    return cfg


def load_server(settings_module=DEFAULT_MODULE) -> ServerSettings:
    try:
        mod = importlib.import_module(settings_module)
    except ModuleNotFoundError:
        mod = importlib.import_module(GLOBAL_MODULE)
    settings_dict = {}
    for m in dir(mod):
        if m.isupper():
            # sets.add(m)
            value = getattr(mod, m)
            settings_dict[m] = value

    # if settings_dict.get("SECURITY"):
    #    settings_dict["SECURITY"] = json.dumps(settings_dict["SECURITY"])
    cfg = ServerSettings(**settings_dict)
    cfg.SETTINGS_MODULE = settings_module

    if not cfg.DEBUG:
        _level = _get_level(cfg.LOGLEVEL)
    else:
        _level = logging.DEBUG

    # set BASE_PATH

    os.environ[defaults.BASE_PATH_ENV] = cfg.BASE_PATH
    logging.config.dictConfig(cfg.LOGCONFIG)
    log = logging.getLogger("nbwork.server")
    log.debug("Using {cfg.SETTINGS_MODULE} as config module")

    return cfg



from typing import List

from sanic import Blueprint, Request
from sanic.response import empty, json, stream
from sanic_ext import openapi

from labfunctions.defaults import API_VERSION
from labfunctions.managers import runtimes_mg
from labfunctions.security.web import protected
from labfunctions.types.runtimes import RuntimeData, RuntimeReq
from labfunctions.web.utils import get_query_param2

runtimes_bp = Blueprint("runtimes", url_prefix="runtimes", version=API_VERSION)


@runtimes_bp.get("/<projectid>")
@openapi.parameter("projectid", str, "path")
@openapi.parameter("lt", int, "query")
@openapi.response(200, List[RuntimeData], "Collection of runtimes")
@protected()
async def runtimes_list(request: Request, projectid: str):
    """get a list of runtimes, by default they are ordered by created_at field"""
    lt = get_query_param2(request, "lt", 10)
    session = request.ctx.session
    async with session.begin():
        rows = await runtimes_mg.get_list(session, projectid, limit=lt)

    return json([r.dict() for r in rows], 200)


@runtimes_bp.post("/<projectid>")
@openapi.body({"application/json": RuntimeReq})
@openapi.parameter("projectid", str, "path")
@openapi.response(200)
@openapi.response(201)
@protected()
async def runtimes_create(request: Request, projectid):
    session = request.ctx.session
    rq = RuntimeReq(**request.json)
    async with session.begin():
        created = await runtimes_mg.create(session, rq)
    code = 201
    if not created:
        code = 200
    return json({"msg": "ok"}, code)


@runtimes_bp.delete("/<projectid>/<rid: str>")
@openapi.parameter("projectid", str, "path")
@openapi.parameter("id", str, "path")
@openapi.response(200)
@protected()
async def runtimes_delete(request: Request, projectid: str, rid: str):
    session = request.ctx.session
    async with session.begin():
        await runtimes_mg.delete_by_rid(session, rid)

    return json({"msg": "ok"}, 200)



class RuntimeCreationError(Exception):
    def __init__(self, docker_name, projectid):
        msg = f"Error with the creation of runtime {docker_name} for {projectid}"
        super().__init__(msg)


class RuntimeNotFound(Exception):
    def __init__(self, runtimeid):
        _msg = f"runtimeid: {runtimeid} not found"
        super().__init__(_msg)



import json
import logging
import os
from datetime import datetime, timedelta

from labfunctions import client, defaults, secrets

# from labfunctions.executors import context
# from labfunctions.conf.server_settings import settings
from labfunctions.types import ExecutionNBTask, ExecutionResult

from .nbtask_base import NBTaskDocker


def docker_exec(ctx: ExecutionNBTask) -> ExecutionResult:
    """
    It will get a wfid from the control plane.
    This function runs in RQ Worker from a data plane machine.

    Passing task exec information though serialized as an environment
    variable could be a questionable design decision, checks:
        - https://www.in-ulm.de/~mascheck/various/argmax/
        - and https://stackoverflow.com/questions/1078031/what-is-the-maximum-size-of-a-linux-environment-variable-value
        - and getconf -a | grep ARG_MAX # (value in kib)
    """

    agent_token = os.getenv("NB_AGENT_TOKEN")
    refresh_token = os.getenv("NB_AGENT_REFRESH_TOKEN")
    url_service = os.getenv("NB_WORKFLOW_SERVICE")
    nbclient = client.agent(
        url_service=url_service,
        token=agent_token,
        refresh=refresh_token,
        projectid=ctx.projectid,
    )
    runner = NBTaskDocker(nbclient)
    result = runner.run(ctx)
    if result.error and not os.getenv("DEBUG"):
        runner.register(result)
    return result
    # error_result = context.make_error_result(exec_ctx, elapsed)
    # ag_client.history_register(result)



import json
from typing import List, Union

import aioredis

from labfunctions.types.events import EventSSE
from labfunctions.utils import secure_filename

BLOCK_MS = 15 * 1000
DEFAULT_TTL = 60 * 60


class EventManager:
    def __init__(
        self, redis: aioredis.client.Redis, block_ms=BLOCK_MS, ttl_secs=DEFAULT_TTL
    ):
        """
        It manages the interaction with Redis Streams data structure
        Also it format data into SSE format.

        :param redis: A redis instance
        :param block_ms: wait x time in milliseconds for a new message
        in the stream channel
        :param ttl_secs: because each stream should be temporal for each execution,
        it should be deleted after `ttl_secs`
        """
        self.redis = redis
        self._block_ms = block_ms
        self._ttl = ttl_secs

    async def read(self, channel, msg_id, block_ms=None) -> Union[List[EventSSE], None]:

        block_ms = block_ms or self._block_ms
        stream = {channel: msg_id}
        rsp = await self.redis.xread(stream, block=block_ms)
        if rsp:
            events = []
            for msg in rsp[0][1]:
                _id = msg[0]
                _data = msg[1]
                # _jdata = json.dumps(_data)
                _evt = EventSSE(id=_id, data=_data["msg"], event=_data["event"])

                events.append(_evt)

            return events
        return None

    async def publish(self, channel, evt: EventSSE, ttl_secs=None):
        ttl_secs = ttl_secs or self._ttl

        async with self.redis.pipeline() as pipe:
            event_type = evt.event or ""
            res = (
                await pipe.xadd(channel, fields={"msg": evt.data, "event": event_type})
                .expire(channel, ttl_secs)
                .execute()
            )
        return res

    @staticmethod
    def generate_channel(projectid, execid):
        pid = secure_filename(projectid)
        eid = secure_filename(execid)
        return f"{pid}.{eid}"

    @staticmethod
    def format_sse(evt: EventSSE) -> str:
        """
        It transforms an EventSSE object into a compliant EventSource string
        https://developer.mozilla.org/en-US/docs/Web/API/Server-sent_events/Using_server-sent_events#event_stream_format
        https://maxhalford.github.io/blog/flask-sse-no-deps/
        """

        msg = f"data: {evt.data}\n\n"
        if evt.event is not None:
            msg = f"event: {evt.event}\n{msg}"
        if evt.id is not None:
            msg = f"id: {evt.id}\n{msg}"
        return msg

    @staticmethod
    def from_sse2event(msg: str) -> EventSSE:
        event = None
        id = None
        data = None
        for line in msg.split("\n"):
            if line.startswith("data:"):
                data = line.split(":", maxsplit=1)[1].strip()
            elif line.startswith("event:"):
                event = line.split(":", maxsplit=1)[1].strip()
            elif line.startswith("id:"):
                id = line.split(":", maxsplit=1)[1].strip()
        return EventSSE(id=id, event=event, data=data)
